/*! ********************************************
 * MultiTimberSynthBase.as
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
package sora.graybox.synth 
{
import flash.utils.ByteArray;
import sora.graybox.GrayBox;
import sora.graybox.IGrayBoxModule;
import sora.graybox.MidiEvent;
import sora.graybox.ProcessData;

public class MultiTimberSynthBase implements IGrayBoxModule
{
// Factory -----
	protected function _recreateVU(maxPoly:int):void{}
	protected function _resetKeyParams():void{}
// Tables -----
	protected const _MASTER_VOLUME_TABLE:Vector.<Number> = new Vector.<Number>(128, true);  // for sysEx, NRPN
	protected function _resetVolumeTables():void
	{
		for (var i:int = 0; i < 128; i++)
		{
			// 0 to -96dB
			_MASTER_VOLUME_TABLE[i] = Math.pow(2.0, (128 - i) / 8.0) / 65536.0;
		}
	}
// Synthesizer - Data -----
	public function factoryDataReset():void{}
// Synthesizer - Main -----
	// Synth Status
	public var midiInPorts  :int = 0xFFFF;
	public var midiOutPorts :int = 0xFFFF;
	public var receiveSysEx :Boolean = false;
	protected var _sampleRate     :Number = 44100.0;
	protected var _referencePitch :Number = 440.0;
	protected var _masterVolume   :Number = 1.0;
	// Channel Paramters
	public const VOICE_PARTS :Vector.<VoicePart> = new Vector.<VoicePart>(16, true);
	// VoiceUnits
	protected var _maxPolyphony :int = 0;
	protected var _voiceUnits   :Vector.<VoiceUnitBase> = null;
	protected var _nton         :int = 0;
	protected var _ntonM        :int = 0;
	// getter, setter ---
	public function getMasterVolume():Number { return _masterVolume; }
	public function setMasterVolume(val:Number):void
	{ 
		_masterVolume = val;
		for (var i:int = 0; i < _maxPolyphony; i++)_voiceUnits[i].setMasterVolume(_masterVolume);
	}
	public function getSampleRate():Number{ return _sampleRate; }
	public function setSampleRate(val:Number):void{ _sampleRate = val; _resetKeyParams(); }
	public function getReferencePitch():Number{ return _referencePitch; }
	public function setReferencePitch(val:Number):void{ _referencePitch = val; _resetKeyParams(); }
	public function getPolyphony():int
	{
		var i:int = 0, _cnt:int = 0;
		for ( ; i < _maxPolyphony; i++) if (_voiceUnits[i].state)_cnt++;
		return _cnt;
	}
	public function getMaxPolyphony():int{ return _maxPolyphony; }
	public function setMaxPolyphony(num:int):int
	{
		if (num < 1 || num == _maxPolyphony) return _maxPolyphony;
		if (num >= 1024) num = 1024;
		else for (var v:int = 1; v < 1024; v <<= 1) if (v >= num){ num = v; break; } 
		_nton = 0;
		_ntonM = num - 1;
		_recreateVU(num);
		_resetKeyParams();
		return _maxPolyphony;
	}
// Constructor
	public function MultiTimberSynthBase(maxPoly:int = 32, refPitch:Number = 440.0, masVol:Number = 1.0):void
	{
		_referencePitch = refPitch;
		_masterVolume = masVol;
	// Table ---
		_resetVolumeTables();
	// RPN ---
		var i:int;
		for (i = 0; i < 256; i++)_RPN_FUNC[i] = _rpnFunc_UB;
		_RPN_FUNC[0] = _rpnFunc_00;
		_RPN_FUNC[1] = _rpnFunc_01;
		_RPN_FUNC[2] = _rpnFunc_02;
		_RPN_FUNC[128] = _nrpnFunc_00;
		_RPN_FUNC[256] = _rpnFunc_Null;
	// CC ---
		for (i =   0; i < 120; i++)_CC_FUNC[i] = _ccFunc_channelVoiceUB;
		for (i = 120; i < 128; i++)_CC_FUNC[i] = _ccFunc_channelModeUB;
		_CC_FUNC[0x01] = _ccFunc_01;
		_CC_FUNC[0x06] = _ccFunc_06;
		_CC_FUNC[0x07] = _ccFunc_07;
		_CC_FUNC[0x0A] = _ccFunc_0A;
		_CC_FUNC[0x0B] = _ccFunc_0B;
		_CC_FUNC[0x26] = _ccFunc_26;
		_CC_FUNC[0x60] = _ccFunc_60;
		_CC_FUNC[0x61] = _ccFunc_61;
		_CC_FUNC[0x62] = _ccFunc_62;
		_CC_FUNC[0x63] = _ccFunc_63;
		_CC_FUNC[0x64] = _ccFunc_64;
		_CC_FUNC[0x65] = _ccFunc_65;
		_CC_FUNC[0x78] = _ccFunc_78;
		_CC_FUNC[0x79] = _ccFunc_79;
		_CC_FUNC[0x7B] = _ccFunc_7B;
	// Part ---
		for (i = 0; i < 16; i++) VOICE_PARTS[i] = new VoicePart(1);
	// VoiceUnit ---
		if (maxPoly >= 1024) maxPoly = 1024;
		else for (var v:int = 1; v < 1024; v <<= 1) if (v >= maxPoly){ maxPoly = v; break; } 
		_nton = 0;
		_ntonM = maxPoly - 1;
		_maxPolyphony = maxPoly;
	}
// Controls ---
	public function initialize():void
	{
		_resetVolumeTables();
		allSoundStop();
		allPartReset(true);
		factoryDataReset();
		_rpnCtrlFlag = true;
		_nrpnCtrlFlag = false;
		receiveSysEx = false;
	}
	public function getDefaultPartMode(part:int):int { return VOICE_PARTS[part].defaultMode; }
	public function setDefaultPartMode(part:int, mode:int):void { VOICE_PARTS[part].defaultMode = mode; }
	public function getPartMode(part:int):int { return VOICE_PARTS[part].mode; }
	public function setPartMode(part:int, mode:int):void
	{
		VOICE_PARTS[part].mode = mode;
		var i:int, _vu:VoiceUnitBase;
		for (i = 0; i < _maxPolyphony; i++)
		{
			_vu = _voiceUnits[i];
			if ( _vu.partNumber == part)_vu.soundOff();
		}
	}
	public function allPartReset(modeReset:Boolean = false):void
	{
		for (var i:int = 0; i < 16; i++) VOICE_PARTS[i].reset();
		if (modeReset)for (i = 0; i < 16; i++)VOICE_PARTS[i].mode = VOICE_PARTS[i].defaultMode;
	}
	public function allSoundStop():void
	{
		for (var i:int = 0; i < _maxPolyphony; i++)_voiceUnits[i].killProcess();
	}
	public function allSoundOff():void
	{
		for (var i:int = 0; i < _maxPolyphony; i++)_voiceUnits[i].soundOff();
	}
// MidiCtrl ---
	protected var _evBuff:MidiEvent = new MidiEvent();
	public function noteOff(ch:int, key:int):void
	{
		//_evBuff.type = 0x80;
		_evBuff.channel = ch;
		_evBuff.value0 = key;
		_noteOff(_evBuff);
	}
	public function noteOn(ch:int, key:int, velo:int):void
	{
		if (VOICE_PARTS[ch].mode == 0) return;
		//_evBuff.type = 0x90;
		_evBuff.channel = ch;
		_evBuff.value0 = key;
		_evBuff.value1 = velo;
		_noteOn(_evBuff);
	}
	public function keyAfterTouch(ch:int, key:int, val:int):void
	{
		//_evBuff.type = 0xA0;
		_evBuff.channel = ch;
		_evBuff.value0 = key;
		_evBuff.value1 = val;
		_keyAfterTouch(_evBuff);
	}
	public function controleChange(ch:int, cc:int, val:int):void
	{
		//_evBuff.type = 0xB0;
		_evBuff.channel = ch;
		//_evBuff.value0 = cc;
		_evBuff.value1 = val;
		_CC_FUNC[cc](_evBuff);
	}
	public function programChange(ch:int, num:int):void
	{
		//_evBuff.type = 0xC0;
		_evBuff.channel = ch;
		_evBuff.value0 = num;
		_programChange(_evBuff);
	}
	public function channelAfterTouch(ch:int, val:int):void
	{
		//_evBuff.type = 0xD0;
		_evBuff.channel = ch;
		_evBuff.value0 = val;
		_channelAfterTouch(_evBuff);
	}
	public function pitchBendChange(ch:int, lsb:int, msb:int):void
	{
		//_evBuff.type = 0xE0;
		_evBuff.channel = ch;
		_evBuff.value0 = lsb;
		_evBuff.value1 = msb;
		_pitchBendChange(_evBuff);
	}
	// val[0 - 16383]
	public function pitchBendChangeW(ch:int, val:int):void
	{
		//_evBuff.type = 0xE0;
		_evBuff.channel = ch;
		_evBuff.value0 = val & 0x7f;
		_evBuff.value1 = val >> 7 & 0x7f;
		_pitchBendChange(_evBuff);
	}
	public function sysExMessage(msg:ByteArray):void
	{
		if (!receiveSysEx) return;
		//_evBuff.type = 0xF0;
		_evBuff.channel = 0;
		_evBuff.data = msg;
		_readSysExData(_evBuff);
		_evBuff.data = null;
	}
// Interface - IfGrayBoxModule -----
	public function updateSampleRate(sampleRate:Number):void
	{
		if (sampleRate == _sampleRate) return;
		setSampleRate(sampleRate);
	}
	public function updateAudioChannels(ch:int):void{}
	public function updateProcessBlockSize(size:int):void{}
	
	public function resetModule():void
	{
		allSoundStop();
		allPartReset();
	}
	public function initiModule():void{ initialize(); }
	public function connectModule(host:GrayBox):void{ setSampleRate(host.getSampleRate()); }
	public function disconnectModule():void{ allSoundStop(); }
	public function updateBypassStatus(bypass:Boolean):void{ allSoundStop(); }
	
	public function startProcess():void{ allSoundStop(); }
	public function stopProcess():void{ allSoundStop(); }
	public function suspendProcess():void{ allSoundStop(); }
	public function resumeProcess():void{ allSoundStop(); }
	public function process(procData:ProcessData):void
	{
		var i:int, _events:MidiEvent = procData.midiEvents.next,
			_outL:Vector.<Number> = procData.audioOut[0],
			_outR:Vector.<Number> = procData.audioOut[1],
			_pos:int = 0, _startFrame:int = procData.offsetFrames, _endFrame:int,
			_vu:Vector.<VoiceUnitBase> = _voiceUnits, _mxp:int = _maxPolyphony;
		// Prepare Process -->
		for (i = 0; i < _mxp; i++)_vu[i].startFrame = _startFrame;
		// Process Events -->
		while (_events)
		{
			if ((_events.midiPorts & midiInPorts) == 0){ _events = _events.next; continue; }
			switch (_events.type)
			{
				case 0x80:
					if (_events.offsetFrames > _pos)
					{
						_pos = _events.offsetFrames;
						_endFrame = _startFrame + _pos;
						for (i = 0; i < _mxp; i++)if (_vu[i].state)_vu[i].process(_outL, _outR, _endFrame);
					}
					_noteOff(_events);
					break;
				case 0x90:
					if (VOICE_PARTS[_events.channel].mode == 0) break;
					if (_events.offsetFrames > _pos)
					{
						_pos = _events.offsetFrames;
						_endFrame = _startFrame + _pos;
						for (i = 0; i < _mxp; i++)if (_vu[i].state)_vu[i].process(_outL, _outR, _endFrame);
					}
					_noteOn(_events);
					break;
				case 0xA0:
					if (_events.offsetFrames > _pos)
					{
						_pos = _events.offsetFrames;
						_endFrame = _startFrame + _pos;
						for (i = 0; i < _mxp; i++)if (_vu[i].state)_vu[i].process(_outL, _outR, _endFrame);
					}
					_keyAfterTouch(_events);
					break;
				case 0xB0:
					if (_events.offsetFrames > _pos)
					{
						_pos = _events.offsetFrames;
						_endFrame = _startFrame + _pos;
						for (i = 0; i < _mxp; i++)if (_vu[i].state)_vu[i].process(_outL, _outR, _endFrame);
					}
					_CC_FUNC[_events.value0](_events);
					break;
				case 0xC0:
					if (_events.offsetFrames > _pos)
					{
						_pos = _events.offsetFrames;
						_endFrame = _startFrame + _pos;
						for (i = 0; i < _mxp; i++)if (_vu[i].state)_vu[i].process(_outL, _outR, _endFrame);
					}
					VOICE_PARTS[_events.channel].programChange = _events.value0;
					break;
				case 0xD0:
					if (_events.offsetFrames > _pos)
					{
						_pos = _events.offsetFrames;
						_endFrame = _startFrame + _pos;
						for (i = 0; i < _mxp; i++)if (_vu[i].state)_vu[i].process(_outL, _outR, _endFrame);
					}
					_channelAfterTouch(_events);
					break;
				case 0xE0:
					if (_events.offsetFrames > _pos)
					{
						_pos = _events.offsetFrames;
						_endFrame = _startFrame + _pos;
						for (i = 0; i < _mxp; i++)if (_vu[i].state)_vu[i].process(_outL, _outR, _endFrame);
					}
					_pitchBendChange(_events);
					break;
				case 0xF0:
					if (!receiveSysEx) break;
					if (_events.offsetFrames > _pos)
					{
						_pos = _events.offsetFrames;
						_endFrame = _startFrame + _pos;
						for (i = 0; i < _mxp; i++)if (_vu[i].state)_vu[i].process(_outL, _outR, _endFrame);
					}
					_readSysExData(_events);
					break;
			}
			_events = _events.next;
		}
		// End Process -->
		_endFrame = _startFrame + procData.sampleFrames;
		for (i = 0; i < _mxp; i++)if (_vu[i].state)_vu[i].process(_outL, _outR, _endFrame);
	}
// MidiEvent ---
	protected function _noteOff(ev:MidiEvent):void
	{
		var i:int = 0, _key:int = ev.value0, _ch:int = ev.channel, _vu:VoiceUnitBase;
		for ( ; i < _maxPolyphony; i++)
		{
			_vu = _voiceUnits[i];
			// KeyOn -->
			if (_vu.state & 0x010) if (_vu.partNumber == _ch && _vu.noteNumber == _key)_vu.keyOff();
		}
	}
	protected function _noteOn(ev:MidiEvent):void
	{
		var i:int, _vu:VoiceUnitBase;
		for (i = 0; i < _maxPolyphony; i++)
		{
			_vu = _voiceUnits[i];
			if (_vu.state == 0)
			{
				_vu.startFrame = ev.offsetFrames;
				_vu.keyOn(ev.channel, ev.value0, ev.value1);
				return;
			}
		}
		var _len:int = _nton + _maxPolyphony;
		for (i = _nton; i < _len; i++)
		{
			_vu = _voiceUnits[i & _ntonM];
			// (KeyOff || SoundOff) -->
			if (_vu.state & 0x60)
			{
				_nton = _nton + 1 & _ntonM;
				_vu.startFrame = ev.offsetFrames;
				_vu.keyOn(ev.channel, ev.value0, ev.value1);
				return;
			}
		}
		for (i = _nton; i < _len; i++)
		{
			_vu = _voiceUnits[i & _ntonM];
			// KeyOn -->
			if (_vu.state & 0x10)
			{
				_nton = _nton + 1 & _ntonM;
				_vu.startFrame = ev.offsetFrames;
				_vu.keyOn(ev.channel, ev.value0, ev.value1);
				return;
			}
		}
	}
	protected function _keyAfterTouch(ev:MidiEvent):void{}
	protected function _programChange(ev:MidiEvent):void
	{
		var _ch:int = ev.channel;
		VOICE_PARTS[_ch].programChange = ev.value0;
		if (VOICE_PARTS[_ch].mode == 0) return;
		for (var i:int = 0; i < _maxPolyphony; i++) if (_voiceUnits[i].partNumber == _ch)_voiceUnits[i].soundOff();
	}
	protected function _channelAfterTouch(ev:MidiEvent):void
	{
		VOICE_PARTS[ev.channel].chPressure = ev.value0;
	}
	protected function _pitchBendChange(ev:MidiEvent):void
	{
		var i:int = 0, _ch:int = ev.channel,
			_pt:VoicePart = VOICE_PARTS[_ch], _vu:VoiceUnitBase;
		_pt.pitchBend = ev.value1 << 7 | ev.value0;
		if (_pt.mode == 0) return;
		for ( ; i < _maxPolyphony; i++)
		{
			_vu = _voiceUnits[i];
			if (_vu.partNumber == _ch)_vu.updPartPitchBend(_pt.pitchBend);
		}
	}
	protected function _readSysExData(ev:MidiEvent):void {}
// MidiEvent - CC -----
	// Function Detail : function(ev:MidiEvent):void
	protected const _CC_FUNC :Vector.<Function> = new Vector.<Function>(128, true);
	private function _ccFunc_channelVoiceUB(ev:MidiEvent):void
	{
		VOICE_PARTS[ev.channel].ccValues[ev.value0] = ev.value1;
	}
	private function _ccFunc_channelModeUB(ev:MidiEvent):void{}
	private function _ccFunc_01(ev:MidiEvent):void
	{
		// Modulation --->
		var i:int, _ch:int = ev.channel, 
			_pt:VoicePart = VOICE_PARTS[_ch], _vu:VoiceUnitBase;
		_pt.ccValues[1] = ev.value1;
		if (_pt.mode == 0) return;
		for (i = 0; i < _maxPolyphony; i++)
		{
			_vu = _voiceUnits[i];
			if (_vu.partNumber == _ch)_vu.updPartModulation(ev.value1);
		}
	}
	private function _ccFunc_06(ev:MidiEvent):void
	{
		// Data Entry MSB --->
		VOICE_PARTS[ev.channel].ccValues[6] = ev.value1;
		_RPN_FUNC[VOICE_PARTS[ev.channel].paramNumber](ev.channel, ev.value1 << 7, 0x7f, 0);
	}
	private function _ccFunc_07(ev:MidiEvent):void
	{
		// Volume --->
		var i:int, _ch:int = ev.channel, 
			_pt:VoicePart = VOICE_PARTS[_ch], _vu:VoiceUnitBase;
		_pt.ccValues[7] = ev.value1;
		if (_pt.mode == 0) return;
		for (i = 0; i < _maxPolyphony; i++)
		{
			_vu = _voiceUnits[i];
			if (_vu.partNumber == _ch)_vu.updPartVolume(ev.value1);
		}
	}
	private function _ccFunc_0A(ev:MidiEvent):void
	{
		// Pan --->
		var i:int, _ch:int = ev.channel, 
			_pt:VoicePart = VOICE_PARTS[_ch], _vu:VoiceUnitBase;
		_pt.ccValues[10] = ev.value1;
		if (_pt.mode == 0) return;
		for (i = 0; i < _maxPolyphony; i++)
		{
			_vu = _voiceUnits[i];
			if (_vu.partNumber == _ch)_vu.updPartPan(ev.value1);
		}
	}
	private function _ccFunc_0B(ev:MidiEvent):void
	{
		// Expression --->
		var i:int, _ch:int = ev.channel, 
			_pt:VoicePart = VOICE_PARTS[_ch], _vu:VoiceUnitBase;
		_pt.ccValues[11] = ev.value1;
		if (_pt.mode == 0) return;
		for (i = 0; i < _maxPolyphony; i++)
		{
			_vu = _voiceUnits[i];
			if (_vu.partNumber == _ch)_vu.updPartExpression(ev.value1);
		}
	}
	private function _ccFunc_26(ev:MidiEvent):void
	{
		// Data Entry LSB --->
		VOICE_PARTS[ev.channel].ccValues[38] = ev.value1;
		_RPN_FUNC[VOICE_PARTS[ev.channel].paramNumber](ev.channel, ev.value1, 0x3f80, 0);
	}
	private function _ccFunc_60(ev:MidiEvent):void
	{
		// Data Inc --->
		_RPN_FUNC[VOICE_PARTS[ev.channel].paramNumber](ev.channel, 0, 0x3fff, 1);
	}
	private function _ccFunc_61(ev:MidiEvent):void
	{
		// Data Dec --->
		_RPN_FUNC[VOICE_PARTS[ev.channel].paramNumber](ev.channel, 0, 0x3fff, -1);
	}
	private function _ccFunc_62(ev:MidiEvent):void
	{
		// NRPN LSB --->
		if (!_nrpnCtrlFlag) return;
		var _pt:VoicePart = VOICE_PARTS[ev.channel];
		_pt.ccValues[98] = ev.value1;
		_pt.nrpn = 0x4000 | _pt.nrpn & 0xff80 | ev.value1;
		if (_pt.nrpn >= 0xc000)  // Received MSB,LSB
		{
			_pt.paramNumber = _pt.nrpn & 0x3fff;
			if (_pt.paramNumber <= 0x7f)_pt.paramNumber |= 0x80;
			else _pt.paramNumber = 0x100;
		}
	}
	private function _ccFunc_63(ev:MidiEvent):void
	{
		// NRPN MSB --->
		if (!_nrpnCtrlFlag) return;
		var _pt:VoicePart = VOICE_PARTS[ev.channel];
		_pt.ccValues[99] = ev.value1;
		_pt.nrpn = 0x8000 | _pt.nrpn & 0xc07f | ev.value1 << 7;
		if (_pt.nrpn >= 0xc000)  // Received MSB,LSB
		{
			_pt.paramNumber = _pt.nrpn & 0x3fff;
			if (_pt.paramNumber <= 0x7f)_pt.paramNumber |= 0x80;
			else _pt.paramNumber = 0x100;
		}
	}
	private function _ccFunc_64(ev:MidiEvent):void
	{
		// RPN LSB --->
		if (!_rpnCtrlFlag) return;
		var _pt:VoicePart = VOICE_PARTS[ev.channel];
		_pt.ccValues[100] = ev.value1;
		_pt.rpn = 0x4000 | _pt.rpn & 0xff80 | ev.value1;
		if (_pt.rpn >= 0xc000)  // Received MSB,LSB
		{
			_pt.paramNumber = _pt.rpn & 0x3fff;
			if (_pt.paramNumber > 0x7f)_pt.paramNumber = 0x100;
		}
	}
	private function _ccFunc_65(ev:MidiEvent):void
	{
		// RPN MSB --->
		if (!_rpnCtrlFlag) return;
		var _pt:VoicePart = VOICE_PARTS[ev.channel];
		_pt.ccValues[101] = ev.value1;
		_pt.rpn = 0x8000 | _pt.rpn & 0xc07f | ev.value1 << 7;
		if (_pt.rpn >= 0xc000)  // Received MSB,LSB
		{
			_pt.paramNumber = _pt.rpn & 0x3fff;
			if (_pt.paramNumber > 0x7f)_pt.paramNumber = 0x100;
		}
	}
	private function _ccFunc_78(ev:MidiEvent):void
	{
		// All Sound Off --->
		var i:int, _ch:int = ev.channel, _vu:VoiceUnitBase;
		if (VOICE_PARTS[ev.channel].mode == 0) return;
		for (i = 0; i < _maxPolyphony; i++)
		{
			_vu = _voiceUnits[i];
			if ( _vu.partNumber == _ch)_vu.soundOff();
		}
	}
	private function _ccFunc_79(ev:MidiEvent):void
	{
		// Reset All Controller --->
		var i:int, _ch:int = ev.channel, 
			_pt:VoicePart = VOICE_PARTS[_ch], _vu:VoiceUnitBase;
		_pt.resetAllControlle();
		if (_pt.mode == 0) return;
		for (i = 0; i < _maxPolyphony; i++)
		{
			_vu = _voiceUnits[i];
			if (_vu.partNumber == _ch)
			{
				_vu.updPartPitchBend(_pt.pitchBend);
				_vu.updPartModulation(_pt.ccValues[1]);
				_vu.updPartExpression(_pt.ccValues[11]);
			}
		}
	}
	private function _ccFunc_7B(ev:MidiEvent):void
	{
		// All Note Off --->
		var i:int, _ch:int = ev.channel, _vu:VoiceUnitBase;
		if (VOICE_PARTS[_ch].mode == 0) return;
		for (i = 0; i < _maxPolyphony; i++)
		{
			_vu = _voiceUnits[i];
			// KeyOn -->
			if ((_vu.state & 0x010) && _vu.partNumber == _ch)_vu.keyOff();
		}
	}
// RPN,NRPN -----
	/*
	- RPN,NRPN -
		         [FE DCBA987 6543210]
		value -> [dc bbbbbbb aaaaaaa]
		a = RPN,NRPN (LSB)
		b = RPN,NRPN (MSB)
		c = Flag (Received LSB)
		d = Flag (Received MSB)
			
		paramNumber -> RPN[0-127], NRPN[128-255], nullRPN[256]
		value <= 100h [100h : null RPN] 
		
		_RPN_FUNC
		Function Detail : function(ch:int, param:int, mask:int, inc:int):void
		
	*/
	protected const _RPN_FUNC     :Vector.<Function> = new Vector.<Function>(257, true);  // RPN[0-127], NRPN[128-255]
	protected var   _rpnCtrlFlag  :Boolean = true;
	protected var   _nrpnCtrlFlag :Boolean = false;
	
	public function getRPNCtrlFlag():Boolean{ return _rpnCtrlFlag; }
	public function setRPNCtrlFlag(flag:Boolean):void
	{
		_rpnCtrlFlag = flag;
		for (var i:int = 0; i < 16; i++) VOICE_PARTS[i].resetRPNControlle();
	}
	public function getNRPNCtrlFlag():Boolean{ return _nrpnCtrlFlag; }
	public function setNRPNCtrlFlag(flag:Boolean):void
	{
		_nrpnCtrlFlag = flag;
		for (var i:int = 0; i < 16; i++) VOICE_PARTS[i].resetRPNControlle();
	}
	private function _rpnFunc_UB(ch:int, param:int, mask:int, inc:int):void{}
	private function _rpnFunc_Null(ch:int, param:int, mask:int, inc:int):void{}
	// bendSens
	private function _rpnFunc_00(ch:int, param:int, mask:int, inc:int):void
	{
		if (mask == 0x3f80) return;  // DataEntry LSB
		// DataEntry MSB, DataInc, DataDec ->
		if (VOICE_PARTS[ch].mode == 0) return;
		var _val:int = (VOICE_PARTS[ch].bendSens << 7) & mask | param;
		_val = (_val >> 7) + inc;
		if		(_val <   0) _val = 0; 
		else if (_val > 127) _val = 127; 
		VOICE_PARTS[ch].bendSens = _val;
		var i:int = 0, _vu:VoiceUnitBase;
		for ( ; i < _maxPolyphony; i++)
		{
			_vu = _voiceUnits[i];
			if (_vu.partNumber == ch)_vu.updPartBendSens(_val);
		}
	}
	// masterTune (FineTune)
	private function _rpnFunc_01(ch:int, param:int, mask:int, inc:int):void
	{
		//fineTune = (fineTune & mask | param) + inc;
		//v = (fineTune - 8192) * 0.08333333333333333;  // 0.083 = 8192(1cotv)/12(tones)/8192(+-)
	}
	// masterTune (CoarseTune)
	private function _rpnFunc_02(ch:int, param:int, mask:int, inc:int):void
	{
		// if (mask != 0x7f) return;  // Data Entry MSB only 
		//coarseTune = param >> 7;
		//v = (coarseTune - 64) * 682.6666666666666;  // 682.6 = 8192(1cotv)/12(tones)
	}
	// masterVolume
	private function _nrpnFunc_00(ch:int, param:int, mask:int, inc:int):void
	{
		if (mask != 0x7f) return;  // Data Entry MSB only 
		setMasterVolume(_MASTER_VOLUME_TABLE[param >> 7]);
	}
}
}