/*! ********************************************
 * SampleData.as
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
package sora.graybox.synth 
{
import flash.utils.ByteArray;
public class SampleData 
{
	/*
	    |          |      |        |
		|  /\    /\|    /\|  /\    |
		|\/  \  /  |\  /  |\/  \/--|
		|     \/   | \/   |        |
		|          |      |        |SoundOff
		|          |      |toRelease
		|          |      |LoopEnd
		|          |LoopStart
		|KeyOn	
		
		- sampleFrames -
		samples : [L,R][L,R]...
		frames    [ 0 ][ 1 ]...
	 */
	internal var sampleData   :Vector.<int> = null;  // Indexed Sample Data
	internal var sampleFrames :int = 0;
	internal var channelCount :int = 0;
	private var _sampleRate   :Number = 0.0;
	private var _recPitch     :Number = 0.0;
	// Synth Param
	internal var pgSample :int = 0;  // SampleRate
	internal var pgPitch  :int = 0;  // RecPitch
	internal var pgKey    :int = 0;  // SampleRate - RecPitch
	internal var plbkS    :int = 0;  // PlayBack Start Frame [plbkS >= 0]
	internal var plbkE    :int = 0;  // PlayBack End Frame   [plbkE <= sampleFrames]
	internal var loopE    :int = 0;  // Loop End Frame       [loopE <= plbkE]
	internal var loopF    :int = 0;  // Loop Frames          [val = loopE - lpStart]
	internal var loopFk   :int = 0;  // Loop Frames          * after -> keyOff
// getter, setter
	public function getSampleFrames():int { return sampleFrames; }
	public function getChannelCount():int { return channelCount; }
	public function getSampleRate():Number { return _sampleRate; }
	public function setSampleRate(sampleRate:Number):void
	{ 
		_sampleRate = sampleRate;
		pgSample = (sampleRate <= 1.0)? 0 : int(Math.log(sampleRate) * (SynthData.POW2T_OCTV_RES * 1.442695040888963387) + 0.5);
		pgKey = pgSample - pgPitch;
	}
	public function getRecPitch():Number { return _recPitch; }
	public function setRecPitch(pitch:Number):void
	{ 
		_recPitch = pitch;
		pgPitch = (pitch <= 1.0)? 0 : int(Math.log(pitch) * (SynthData.POW2T_OCTV_RES * 1.442695040888963387) + 0.5);
		pgKey = pgSample - pgPitch;
	}
	public function getPlayBackStartPoint():int { return plbkS; }
	public function getPlayBackEndPoint():int { return plbkE; }
	public function getLoopStartPoint():int { return (loopF > 0)? loopE - loopF : 0; }
	public function getLoopEndPoint():int { return (loopF > 0)? loopE : 0; }
	public function getLoopable():Boolean { return loopF > 0; }
	
	public function clear():void
	{
		_sampleRate = 0.0;
		_recPitch = 0.0;
		sampleFrames = 0;
		channelCount = 0;
		sampleData = null;
		pgSample = 0;
		pgPitch = 0;
		pgKey = 0;
		plbkS = 0;
		plbkE = 0;
		loopE = 0;
		loopF = 0;
		loopFk = 0;
	}
	public function copy(from:SampleData, deepCopy:Boolean = false):void
	{
		sampleFrames = from.sampleFrames;
		channelCount = from.channelCount;
		_sampleRate = from._sampleRate;
		_recPitch = from._recPitch;
		// Synth Param
		pgSample = from.pgSample;
		pgPitch = from.pgPitch;
		pgKey = from.pgKey;
		plbkS = from.plbkS;
		plbkE = from.plbkE;
		loopE = from.loopE;
		loopF = from.loopF;
		loopFk = from.loopFk;
		// wave
		if (deepCopy && from.sampleData != null)
		{
			var i:int, _len:int = from.sampleData.length,
				_source:Vector.<int> = from.sampleData;
			sampleData = new Vector.<int>(_len, true);
			for (i = 0; i < _len; i++) sampleData[i] = _source[i];
		}
		else sampleData = from.sampleData;
	}
	public function setPlaybackRange(start:int, end:int):void
	{
		if (start >= end || end > sampleFrames) return;
		plbkS = start;
		plbkE = end;
		loopE = plbkE;
		loopF = 0;
		loopFk = 0;
	}
	public function resetLoopPoints(start:int = 0, end:int = 0):void
	{
		// loop
		if (start >= plbkS && start < end && end <= plbkE)
		{
			loopE = end;
			loopF = end - start;
			loopFk = (end == plbkE)? loopF : 0;
		}
		// no loop
		else
		{
			loopE = plbkE;
			loopF = 0;
			loopFk = 0;
		}
	}
	public function setPCMSamples(wave:Vector.<Number>, ch:int):void
	{
		clear();
		// ---
		const _LOGE:Number = -(SynthData.LOG2VT_DYN_RES * 1.442695040888963387),
		      _BOADER:Number = 1.0 / (1 << SynthData.LOG2VT_DYN_BITS),
		      _BOTTOM:int = SynthData.LOG2VT_SIZE;
		var i:int, _len:int = wave.length, v:Number,
			_wave:Vector.<int> = new Vector.<int>(_len, true);
		for (i = 0; i < _len; i++)
		{
			v = wave[i];
			// Positive Value
			if      (v >  _BOADER) _wave[i] =  int(Math.log( v) * _LOGE + 0.5) << 1;
			// Negative Value
			else if (v < -_BOADER) _wave[i] = (int(Math.log(-v) * _LOGE + 0.5) << 1) + 1;  // +1 = Nega
			// 0 value
			else _wave[i] = _BOTTOM;
		}
		sampleData = _wave;
		channelCount = ch;
		sampleFrames = _len / ch;
		// pitch
		_sampleRate = 44100.0;
		setSampleRate(_sampleRate);
		// playback
		plbkE = sampleFrames;
		loopE = sampleFrames;
	}
	public function loadPCMData(dat:ByteArray, sampleFrames:int, ch:int, bitRate:int, sign:Boolean, bigEndian:Boolean):void
	{
		clear();
		// ---
		const _LOGE:Number = -(SynthData.LOG2VT_DYN_RES * 1.442695040888963387),
		      _BOADER:Number = 1.0 / (1 << SynthData.LOG2VT_DYN_BITS),
		      _BOTTOM:int = SynthData.LOG2VT_SIZE;
		function _log2VTindex(val:Number):int
		{
			if      (val >  _BOADER) return  int(Math.log( val) * _LOGE + 0.5) << 1;
			else if (val < -_BOADER) return (int(Math.log(-val) * _LOGE + 0.5) << 1) + 1;  // +1 = Nega
			else return _BOTTOM;
		}
		// ---
		var _endi:String = dat.endian, 
			_pos:uint = dat.position;
		dat.endian = (bigEndian)? "bigEndian" : "littleEndian";
		// ---
		var i:int, _len:int = sampleFrames * ch, _sl:Number, _wave:Vector.<int>;
		switch (bitRate)
		{
			default : clear(); return;
			
			case  8:
				_wave = new Vector.<int>(_len, true);
				_sl = 1.0 / 128.0;
				if (sign) for (i = 0; i < _len; i++) _wave[i] = _log2VTindex(dat.readByte() * _sl);
				else      for (i = 0; i < _len; i++) _wave[i] = _log2VTindex(int(dat.readUnsignedByte() - 128) * _sl);
				break;
				
			case 16:
				_wave = new Vector.<int>(_len, true);
				_sl = 1.0 / 32768.0;
				if (sign) for (i = 0; i < _len; i++) _wave[i] = _log2VTindex(dat.readShort() * _sl);
				else      for (i = 0; i < _len; i++) _wave[i] = _log2VTindex(int(dat.readUnsignedShort() - 32768) * _sl);
				break;
				
			case 24:
				_wave = new Vector.<int>(_len, true);
				_sl = 1.0 / 8388608.0;
				var _v:int, _shif1:int = 0, _shif2:int = 8, _shif3:int = 16;
				if (bigEndian){ _shif1 = 16; _shif3 =  0; }
				if (sign)
				{
					_shif1 += 8; _shif2 += 8; _shif3 += 8;
					for (i = 0; i < _len; i++)
					{
						_v = dat.readUnsignedByte() << _shif1 | dat.readUnsignedByte() << _shif2 | dat.readUnsignedByte() << _shif3;
						_wave[i] = _log2VTindex((_v >> 8) * _sl);
					}
				}
				else for (i = 0; i < _len; i++)
				{
					_v = int(dat.readUnsignedByte() << _shif1 | dat.readUnsignedByte() << _shif2 | dat.readUnsignedByte() << _shif3) - 8388608;
					_wave[i] = _log2VTindex(_v * _sl); 
				}
				break;
				
			case 32:
				_wave = new Vector.<int>(_len, true);
				_sl = 1.0 / 2147483648.0;
				if (sign) for (i = 0; i < _len; i++) _wave[i] = _log2VTindex(dat.readInt() * _sl);
				else      for (i = 0; i < _len; i++) _wave[i] = _log2VTindex(int(dat.readUnsignedInt() - 2147483648) * _sl);
				break;
		}
		sampleData = _wave;
		channelCount = ch;
		this.sampleFrames = sampleFrames;
		// pitch
		_sampleRate = 44100.0;
		setSampleRate(_sampleRate);
		// playback
		plbkE = sampleFrames;
		loopE = sampleFrames;
		// ---
		dat.endian = _endi;
		dat.position = _pos;
	}
// Constructor
	public function SampleData():void{}
}

}